
fprintf('\nPress Ctrl-C to stop\n');

symbolrate=31.25;%rate per second that symbols are sent
Fs=11025*2;%sampling rate
carrierfreq=1000;%carrier frequency
numberofsymbolstoxferpercycle=10;
bitspersymbol=2;%one or two
channelmodelname='none';%some channels are 'iturHFMQ';'iturHFMM';'iturHFMD';'iturHFHQ';'iturHFLQ';'iturHFLM';'iturHFHM';'iturHFLD';'iturHFHD'
EbNo=Inf;%in dB. is for the messages that gets encoded through LDPC, use Inf for no noise
sigpower=0.001;%in Watts if over R=1 ohms
useequalizer=false;
soundout=true;
soundin=false;
enable_tx=true;
enable_rx=true;
stop_rx_when_finished_tx=true;
usescope=true;

%--init start

%init varicode
varicode_init;

%init LDPC
ldpc_init;

oversample_factor=14;
numberofbitstoxferpercycle=bitspersymbol*numberofsymbolstoxferpercycle;
[upfactor,downfactor] = rat(Fs/symbolrate);

hap = dsp.AudioPlayer('SampleRate',Fs);
har = dsp.AudioRecorder('SampleRate',Fs);
har.NumChannels=1;
har.SamplesPerFrame=numberofsymbolstoxferpercycle*upfactor/downfactor;


hRrcTx = comm.RaisedCosineTransmitFilter;
hRrcTx.RolloffFactor=0.2;
hRrcTx.FilterSpanInSymbols=10;
hRrcTx.OutputSamplesPerSymbol=upfactor;

hRrcRx = comm.RaisedCosineReceiveFilter;
hRrcRx.RolloffFactor=0.2;
hRrcRx.FilterSpanInSymbols=10;
hRrcRx.InputSamplesPerSymbol=upfactor;
hRrcRx.DecimationFactor=upfactor/oversample_factor;

hPFOu = comm.PhaseFrequencyOffset('FrequencyOffset', carrierfreq,'SampleRate',Fs);
hPFOd = comm.PhaseFrequencyOffset('FrequencyOffset', -carrierfreq,'SampleRate',Fs);

hMod = comm.DPSKModulator('PhaseRotation',0,'BitInput',true,'ModulationOrder',2^bitspersymbol);
hDemod = comm.PSKDemodulator('PhasOffset',0,'BitOutput',true,'ModulationOrder',2^bitspersymbol,'DecisionMethod','Log-likelihood ratio','Variance',0.8);
hDemod_last_diff_bit=0.1;

eqobj = lineareq(100,signlms(0.00001));
%eqobj = lineareq(200,cma(0.00001));
eqobj.SigConst = [1 -1 1i -1i];
eqobj.ResetBeforeFiltering=false;
eqobj.Weights(1)=1;
eqobj.WeightInputs(1)=1;

DELAY = dsp.Delay(467);

hTimingSync = comm.GardnerTimingSynchronizer('SamplesPerSymbol',oversample_factor,'ErrorUpdateGain',0.01);%to small and to slow, to big and fails on fades

agc = comm.AGC;
agc.MaximumGain=1000;
agc.GainOutputPort=true;
agc.DetectorMethod='Square Law';
agc.StepSize=0.12;
agc.UpdatePeriod=10;

if usescope
    hConstellationScope = comm.ConstellationDiagram('XLimits',[-1.5 1.5],'YLimits',[-1.5 1.5]);
    hConstellationScope.ReferenceConstellation=[-1,1];
    hConstellationScope.SymbolsToDisplaySource='Property';
    hConstellationScope.SymbolsToDisplay=32;
    hConstellationScope.ShowReferenceConstellation=false;
    hConstellationScope.show;
    rxsig_bb_low=zeros(numberofsymbolstoxferpercycle,1);
    hConstellationScope.step(complex(rxsig_bb_low));
end

%channel
usechannel=false;
if exist('channelmodelname','var') && ~strcmp(channelmodelname,'none')
    chan = stdchan(1/Fs, 1, channelmodelname);
    chan.NormalizePathGains = 1;
    chan.ResetBeforeFiltering=false;
    usechannel=true;
end

codeoffsetlockoncountdown=0;%lock to stop getting new code locks to fast
demodDataBuffer=zeros(size(hEnc.ParityCheckMatrix,2)+1,1);%unaligned buffer of demod data
pastdemoddata=[];%aligned remainder of codewords
pisum_past_raw=zeros(codewordsz*3/3,1);%inverse parity buffer
pisumpeakbuffer=zeros(3,1);%buffer for finding peak in invese parity


%------ make encoded data to send

%create message
message=[];
for kk=1:3;%100
    message=sprintf('%s   .. OK???? %d This is a longer message as things are sent little by little...     \n',message,kk);
end

%vari encode it
data_org=vari_encode(message);
data_org=[ones(32*3,1);data_org;zeros(32,1)];%some idle bits so dont miss the start of the data. LDPC atm syncs on ones.

%LDPC encode data
msgs=buffer(data_org,msgwordsz);
cwds=zeros(codewordsz,size(msgs,2));
for msgcwd_ptr=1:size(msgs,2)
    cwds(:,msgcwd_ptr)=step(hEnc, msgs(:,msgcwd_ptr));
    cwds(:,msgcwd_ptr)=cwds(idx,msgcwd_ptr);
end
data = reshape(cwds,[numel(cwds),1]);

if enable_tx
    %little bits to send just a little at a time
    databitarray=buffer(data,numberofbitstoxferpercycle);
else
    databitarray=[];
end

%--------- done makeing encoded data to send

sig=zeros(10000,1);
databitarrayp=1;


while enable_rx || enable_tx
    
    %send data and add channel impairments to signal
    if databitarrayp<=size(databitarray,2)
        
        %mod upsample rrc fir downsample
        modsig_bb_low = step(hMod,1-databitarray(:,databitarrayp));
        modsig_bb_ehr = step(hRrcTx,modsig_bb_low);
        modsig_bb_hr = downsample(modsig_bb_ehr,downfactor);
        
        %add hf channel in baseband
        scale = 1/sqrt(mean(abs(modsig_bb_hr).^2));
        if (scale>10000) || isinf(scale)
            scale=1;
        end
        modsig_bb_hr=scale*modsig_bb_hr*sqrt(sigpower);
        if usechannel
            modsig_bb_hr=(filter(chan,modsig_bb_hr));
        end
        
        %raise frequency
        csig=hPFOu.step(modsig_bb_hr);
        sig=sqrt(2)*real(csig);
        
        %calculate variance of noise for AWGN given wanted EbNo
        netbitspersymbol=bitspersymbol*(msgwordsz/codewordsz);
        fb=symbolrate*netbitspersymbol;
        Eb=sigpower/fb;%need to use average power of signal without HF channel,HF channel is normalized so is ok
        No=Eb/(10^(EbNo/10));
        varn=No*(Fs/2);
        noise = normrnd(0,sqrt(varn),numel(sig),1);
        sig=sig+noise;
        
        if max(abs(sig))>1
            error('Signal clipping. Reduce signal power');
        end
        
        %delay. no real reason but can adjust to see what effect delaying
        %the out signal will do
        sig = step(DELAY,sig);
        
        databitarrayp=databitarrayp+1;
    else
        %finished sending data
        sig=zeros(numberofsymbolstoxferpercycle*upfactor/downfactor,1);
        enable_tx=false;
        if stop_rx_when_finished_tx
            enable_rx=false;
        end
    end

    %to sound card
    if soundout
        hap.step(sig);
    end
    
    %from sound card
    if soundin
        sig=har.step();
    end
    
    %receive message from signal
    if enable_rx
        
        %drop frequency
        rxsig_bb_hr=hPFOd.step(sig);
        
        %upsample rrc fir symbol timing downsample
        rxsig_bb_ehr = upsample(rxsig_bb_hr,downfactor);
        rxsig_bb_low_non_sync = step(hRrcRx,rxsig_bb_ehr);
        [rxsig_bb_low_non_sync,agcgain]=agc.step(rxsig_bb_low_non_sync);%agc for hTimingSync
        rxsig_bb_low = step(hTimingSync,rxsig_bb_low_non_sync);
        
        if useequalizer
            rxsig_bb_low = equalize(eqobj,rxsig_bb_low);
        end
        
        %freq offset estimate (this is not to good. should change to somthing better)
        rxsig_bb_low_angle3=unwrap(angle(rxsig_bb_low.^(2^bitspersymbol)))/(2^bitspersymbol);
        rxsig_bb_low_angle2=(angle(rxsig_bb_low.^(2^bitspersymbol)))/(2^bitspersymbol);
        rxsig_bb_low_angle=(rxsig_bb_low_angle3+rxsig_bb_low_angle2)/2;
        p = polyfit(transpose(1:numberofsymbolstoxferpercycle),rxsig_bb_low_angle,1);
        freq_offset=p(1)*symbolrate/(2*pi);
        
        %nudge frequency
        hPFOd.FrequencyOffset=hPFOd.FrequencyOffset-freq_offset/3;
        if (-hPFOd.FrequencyOffset)>(hPFOu.FrequencyOffset+7)
            hPFOd.FrequencyOffset=-(hPFOu.FrequencyOffset+7);
        end
        if (-hPFOd.FrequencyOffset)<(hPFOu.FrequencyOffset-7)
            hPFOd.FrequencyOffset=-(hPFOu.FrequencyOffset-7);
        end
        hPFOd.FrequencyOffset=0.99*hPFOd.FrequencyOffset-0.01*hPFOu.FrequencyOffset;
        
        %correct for frequency changes
        rotations=detrend(polyval(p,transpose(1:numberofsymbolstoxferpercycle)),0);
        rxsig_bb_low=rxsig_bb_low.*exp(-1i*rotations/10);
        
        %differential soft PSK
        denominator=[hDemod_last_diff_bit;rxsig_bb_low(1:end-1)];
        hDemod_last_diff_bit=rxsig_bb_low(end);
        rxsig_bb_low_abs=rxsig_bb_low./(denominator./abs(denominator));
        rxsig_bb_low_abs(isnan(rxsig_bb_low_abs))=0;
        demodData=step(hDemod,-rxsig_bb_low_abs);
        
        if usescope
            hConstellationScope.step(rxsig_bb_low);
        end
        
        %this is pretty cool i think
        %----------find block offset and align if needed
        demodDataBuffer=[demodDataBuffer;demodData];
        bitptr_max=numel(demodData);
        for bitptr=1:bitptr_max
            
            %countdown a lock that inhibits changing of codeword aligment
            codeoffsetlockoncountdown=codeoffsetlockoncountdown-1;
            if codeoffsetlockoncountdown<0
                codeoffsetlockoncountdown=0;
            end
            
            %check parity starting at second position
            data_perm=demodDataBuffer(1+1+bitptr-1:codewordsz+1+bitptr-1);
            [~,PARITY]=hDec2.step(data_perm(idxinv));
            pisum=sum(1-PARITY);
            
            %keep a list of past results
            pisum_past_raw=circshift(pisum_past_raw,-1);
            pisum_past_raw(end)=pisum;
            
            %shift down
            pisum=pisum-mean(pisum_past_raw)-2*std(pisum_past_raw);
            
            %put in tiny peak detector buffer
            pisumpeakbuffer=circshift(pisumpeakbuffer,-1);
            pisumpeakbuffer(end)=pisum;
            
            %if this bit is one we expect to be the codeword start and we
            %get a peak here then reset the countdown
            if codeoffsetlockoncountdown==1 && (pisumpeakbuffer(1))<(pisumpeakbuffer(2)-std(pisum_past_raw)) && (pisumpeakbuffer(3))<(pisumpeakbuffer(2)-std(pisum_past_raw)) && pisumpeakbuffer(2)>0
                codeoffsetlockoncountdown=codewordsz+1;
            end
            
            %if we find a peak above zero then we have found the start of the code
            if (pisumpeakbuffer(1))<pisumpeakbuffer(2) && (pisumpeakbuffer(3))<pisumpeakbuffer(2) && pisumpeakbuffer(2)>0 %&& pisum_past_raw(end-1)>(msgwordsz/2)
                codewordoffset=mod(numel(pastdemoddata)+bitptr-2,codewordsz);%offset
                data_perm=demodDataBuffer(1+bitptr-1:codewordsz+bitptr-1);
                [~,PARITY]=hDec.step(data_perm(idxinv));%check parity starting at last bit position
                if ~sum(PARITY)%it decodes allright?
                    if codewordoffset%a change in codeword offset?
                        if codeoffsetlockoncountdown<=0%are we allowed to change?
                            fprintf('\n---------------------\nGot LDPC sync\n---------------------\n');
                            pastdemoddata=demodDataBuffer(bitptr:codewordsz+1);%fill pastdemoddata for correct alignment
                            break;
                        end
                    else
                        codeoffsetlockoncountdown=2*codewordsz+1;%use two countdowns here
                    end
                end
            end
            
        end
        demodDataBuffer(1:bitptr_max)=[];
        %----------
        
        %decode and print messages
        [msgs,pastdemoddata]=buffer([pastdemoddata;demodData],codewordsz);
        if numel(msgs)~=0
            msgs(isnan(msgs))=0;
            for msgs_ptr=1:size(msgs,2)%might be more than one block to decode
                demodData= double(step(hDec, msgs(idxinv,msgs_ptr)));%decode block
                fprintf('%c',vari_decode(demodData));%vari decode and print
            end
        end
        
    end
    
end



pause(1);


hap.release();
har.release();

fprintf('\nDone\n');
return

